<?php
namespace App\Model\Table;

use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use Cake\Database\Expression\QueryExpression;

/**
 * Messages Model
 *
 * @method \App\Model\Entity\Message get($primaryKey, $options = [])
 * @method \App\Model\Entity\Message newEntity($data = null, array $options = [])
 * @method \App\Model\Entity\Message[] newEntities(array $data, array $options = [])
 * @method \App\Model\Entity\Message|bool save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \App\Model\Entity\Message patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \App\Model\Entity\Message[] patchEntities($entities, array $data, array $options = [])
 * @method \App\Model\Entity\Message findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TimestampBehavior
 */
class MessagesTable extends Table
{

	/**
	 * Initialize method
	 *
	 * @param array $config The configuration for the Table.
	 * @return void
	 */
	public function initialize(array $config)
	{
		parent::initialize($config);

		$this->table('messages');

		$this->addBehavior('Timestamp');
	}

	/**
	 * Default validation rules.
	 *
	 * @param \Cake\Validation\Validator $validator Validator instance.
	 * @return \Cake\Validation\Validator
	 */
	public function validationDefault(Validator $validator)
	{
		$validator
			->integer('id')
			->requirePresence('id', 'create')
			->notEmpty('id');

		$validator
			->numeric('x')
			->requirePresence('x', 'create')
			->notEmpty('x');

		$validator
			->numeric('y')
			->requirePresence('y', 'create')
			->notEmpty('y');

		$validator
			->numeric('z')
			->requirePresence('z', 'create')
			->notEmpty('z');

		$validator
			->requirePresence('type', 'create')
			->notEmpty('type');

		$validator
			->requirePresence('content', 'create')
			->notEmpty('content');

		$validator
			->requirePresence('tags', 'create')
			->notEmpty('tags');

		$validator
			->integer('found_count')
			->requirePresence('found_count', 'create')
			->notEmpty('found_count');

		return $validator;
	}

	/**
	 * Main discovery logic, get messages near requested coordinates and gives fuzzed hints for messages bit farther away
	 * @param  array  $coordinates Array with keys x, y and z
	 * @return array               Array containing arrays of messages and hints
	 */
	public function getMessagesAndHints($coordinates = []) {
		// Get parameters
		$searchedX = $coordinates['x'];
		$searchedY = $coordinates['y'];
		$searchedZ = $coordinates['z'];

		// Setup helpers
		$messages = [];
		$messageIds = [];
		$hints = [];
		$hintMessageIds = [];

		// Get nearby messages
		$result = $this->find()
			->select(['id', 'x', 'y', 'z', 'content', 'type', 'tags', 'found_count'])
			->where([
				'x >=' => $searchedX - 5.0,
				'x <=' => $searchedX + 5.0,
				'y >=' => $searchedY - 5.0,
				'y <=' => $searchedY + 5.0,
				'z >=' => $searchedZ - 5.0,
				'z <=' => $searchedZ + 5.0
			])
			->map(function ($row) {
				// Split comma separated tag list into array
				$row->tags = $this->splitTagString($row->tags);
				return $row;
			})
			->toList();

		foreach ($result as $message) {
			$messageIds[] = $message->id;

			$messages[] = $this->createDynamicMessage($message->content, $message->tags, $message->type);

			if ((int)$message->found_count === 0) {
				$messages[] = $this->createDynamicMessage('You are first one to discover this message.', [], 'text');
			}
		}

		// Get messages farther away
		$query = $this->find()
			->select(['id', 'x', 'y', 'z', 'content', 'type', 'tags', 'found_count'])
			->where([
				'x >=' => $searchedX - 10.0,
				'x <=' => $searchedX + 10.0,
				'y >=' => $searchedY - 10.0,
				'y <=' => $searchedY + 10.0,
				'z >=' => $searchedZ - 10.0,
				'z <=' => $searchedZ + 10.0
			]);

		// Filter out found messages
		if (!empty($messageIds)) {
			$query->where(['id NOT IN' => $messageIds]);
		}

		$result = $query->map(function ($row) {
				// Split comma separated tag list into array
				$row->tags = $this->splitTagString($row->tags);
				return $row;
			})
			->toList();

		// Calculate relative distance and fuzz ~10% to not give away too easy hints
		foreach ($result as $message) {
			$hintMessageIds[] = $message->id;

			$hints[] = [
				'x' => $this->fuzzDistanceHint(
					floatval($message->x - $searchedX) / 10.0,
					10
				),
				'y' => $this->fuzzDistanceHint(
					floatval($message->y - $searchedY) / 10.0,
					10
				),
				'z' => $this->fuzzDistanceHint(
					floatval($message->z - $searchedZ) / 10.0,
					10
				)
			];
		}

		// Update found counts
		if (!empty($messageIds)) {
			$this->incrementFoundCounters($messageIds);
		}

		// Update found counts
		if (!empty($hintMessageIds)) {
			$this->incrementHintCounters($hintMessageIds);
		}

		// Return results
		return [
			'messages' => $messages,
			'hints' => $hints
		];
	}

	/**
	 * Bulk updates found counts for given messages
	 *
	 * @param  array  $messageIds List of message ids to update
	 * @return void
	 */
	public function incrementFoundCounters($messageIds = []) {
		// Try to convert to array if non-array value is given
		if (!is_array($messageIds)) {
			$messageIds = [(int)$messageIds];
			$messageIds = array_filter($messageIds);
		}

		// Skip update if we still have empty id list
		if (empty($messageIds)) {
			return;
		}

		// Update found counts
		try {
			$expression = new QueryExpression('found_count = found_count + 1');
			$this->updateAll(
				[$expression],
				['id IN' => $messageIds]
			);
		} catch (Exception $e) {
			return;
		}
	}

	/**
	 * Bulk updates hint counts for given messages
	 *
	 * @param  array  $messageIds List of message ids to update
	 * @return void
	 */
	public function incrementHintCounters($messageIds = []) {
		// Try to convert to array if non-array value is given
		if (!is_array($messageIds)) {
			$messageIds = [(int)$messageIds];
			$messageIds = array_filter($messageIds);
		}

		// Skip update if we still have empty id list
		if (empty($messageIds)) {
			return;
		}

		// Update hint counts
		try {
			$expression = new QueryExpression('hint_count = hint_count + 1');
			$this->updateAll(
				[$expression],
				['id IN' => $messageIds]
			);
		} catch (Exception $e) {
			return;
		}
	}

	/**
	 * Create data in message format dynamically
	 *
	 * @param  string $content    Message content
	 * @param  array|string $tags Array of message tags, splits and cleans if string is given
	 * @param  string $type       Message type, defaults to 'text'
	 * @return array              Data in message format
	 */
	public function createDynamicMessage($content = '', $tags = [], $type = 'text') {
		if (!is_array($tags)) {
			$tags = $this->splitTagString($tags);
		}

		return [
			'content' => $content,
			'type' => $type,
			'tags' => $tags
		];
	}

	/**
	 * Split comma separeted list of tags into array
	 *
	 * @param  string $tags List of tags as tring
	 * @return array        Array of tags
	 */
	public function splitTagString($tags = '') {
		return array_filter(
			array_map(
				'trim',
				explode(',', $tags)
			)
		);
	}

	/**
	 * Fuzz given relative distance with integer percent amount
	 *
	 * @param  float   $distance Relative distance to fuzz, expects -1.0 to 1.0
	 * @param  integer $amount   Percent amount of max change, default 15%
	 * @return float             Fuzzed value clamped between -0.98 and 0.98
	 */
	public function fuzzDistanceHint($distance = 0.0, $amount = 15) {
		$changePercent = floatval(mt_rand(100 - $amount, 100 + $amount)) / 100.0;
		$result = floatval($distance) * $changePercent;

		return min(max($result, -0.98), 0.98);
	}
}
